# Copyright 2019 Alvaro Bartolome @ alvarobartt in GitHub
# See LICENSE for details.

from pyrtfolio.StockPortfolio import StockPortfolio


def test_stock_portfolio_errors():
    """
    This functions tests the errors that can be raised whenever there is an error while creating a StockPortfolio.
    """

    portfolio = StockPortfolio()

    params = [
        {
            'stock_symbol': ['error'],
            'stock_country': 'spain',
            'purchase_date': '04/01/2018',
            'num_of_shares': 2,
            'cost_per_share': 7.2,
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': ['error'],
            'purchase_date': '04/01/2018',
            'num_of_shares': 2,
            'cost_per_share': 7.2,
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': 'spain',
            'purchase_date': 'error',
            'num_of_shares': 2,
            'cost_per_share': 7.2,
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': 'spain',
            'purchase_date': '04/01/2050',
            'num_of_shares': 2,
            'cost_per_share': 7.2,
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': 'spain',
            'purchase_date': '04/01/2018',
            'num_of_shares': 'error',
            'cost_per_share': 7.2,
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': 'spain',
            'purchase_date': '04/01/2018',
            'num_of_shares': -7,
            'cost_per_share': 7.2,
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': 'spain',
            'purchase_date': '04/01/2018',
            'num_of_shares': 2,
            'cost_per_share': 'error',
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': 'spain',
            'purchase_date': '04/01/2018',
            'num_of_shares': 2,
            'cost_per_share': -7,
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': 'spain',
            'purchase_date': '04/01/2018',
            'num_of_shares': 2,
            'cost_per_share': 7.2,
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': 'error',
            'purchase_date': '04/01/2018',
            'num_of_shares': 2,
            'cost_per_share': 7.2,
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': 'russia',
            'purchase_date': '04/01/2018',
            'num_of_shares': 2,
            'cost_per_share': 7.2,
        },
        {
            'stock_symbol': 'bbva',
            'stock_country': 'spain',
            'purchase_date': '25/12/2018',
            'num_of_shares': 2,
            'cost_per_share': 7.2,
        },
    ]

    for param in params:
        try:
            portfolio.add_stock(stock_symbol=param['stock_symbol'],
                                stock_country=param['stock_country'],
                                purchase_date=param['purchase_date'],
                                num_of_shares=param['num_of_shares'],
                                cost_per_share=param['cost_per_share'])
        except:
            continue


if __name__ == '__main__':
    test_stock_portfolio_errors()



# Copyright 2019 Alvaro Bartolome @ alvarobartt in GitHub
# See LICENSE for details.

from datetime import date

import investpy
import pandas as pd

from pyrtfolio.Stock import Stock


class StockPortfolio(object):
    """ StockPortfolio is the main class of `pyrtfolio` which is going to manage all the introduced stocks.

    This class is the one that contains the stocks information and the one that will be used by the user in order to
    generate a custom portfolio. So on, this function implements the methods to calculate all the values required in a
    basic portfolio and, as already mentioned, the method to add stocks.

    Attributes:
        stocks (:obj:`list`, protected):
            this list contains all the introduced stocks, which will later be used to generate the portfolio.
        stock_objs (:obj:`list`, protected):
            this list contains all the introduced Stock objects, in order to be able to refresh them.
        data (:obj:`pandas.DataFrame`): it is the generated portfolio, once the addition of every stock is validated.

    """

    def __init__(self):
        """ This is the init method of StockPortfolio class which is launched every time the user instances it.

        This method is the init method of this class, StockPortfolio, and its main function is to init all the
        attributes contained in it. Every time this class is instanced, the attributes values are restored and, so on,
        the portfolio's data is lost if existing for that instance. This class does not take any parameters since 
        they are filled once the class is instanced.

        """
        self._stocks = list()
        self._stock_objs = list()
        self.data = None

    def add_stock(self, stock_symbol, stock_country, purchase_date, num_of_shares, cost_per_share):
        """ Method to add a stock to the portfolio.

        This method adds a stock to the custom portfolio data. Some parameters need to be specified for the introduced
        stock such as the purchase date of the shares, the number of shares bought and the price payed (cost) per every
        share. From this data, the portfolio will be created and the specified calculations will be done, so to give
        the user an overview of his/her own portfolio.

        Args:
            stock_symbol (:obj:`str`): symbol of the Stock that is going to be added to the StockPortfolio.
            stock_country (:obj:`str`): country from where the specified stock_symbol is, so to validate it.
            purchase_date (:obj:`str`):
                date when the shares of the introduced stock were bought, formatted as dd/mm/yyyy.
            num_of_shares (:obj:`int`): amount of shares bought of the specified Stock in the specified date.
            cost_per_share (:obj:`float`): price of every share of the Stock in the specified date.

        """
        stock = Stock(stock_symbol, stock_country, purchase_date, num_of_shares, cost_per_share)
        stock.validate()

        if stock.valid is True:
            self._stock_objs.append(stock)

            info = self.__get_stock_info(stock=stock)

            self._stocks.append(info)
            self.data = pd.DataFrame(self._stocks)
        else:
            raise ValueError("ERROR [0001]: The introduced Stock is not valid.")

    def __get_stock_info(self, stock):
        """ Method to get the stock information once it is validated.
        
        This method retrieves the historical data of the introduced Stock in order to get its current 
        price which will be used for the required calculations to generate the StockPortfolio. So on,
        this function is both going to retrieve Stock data and calculate the required values for the 
        StockPortfolio.

        Args:
            stock (:obj:`pyrtfolio.Stock`): Stock object with all its information after validated.

        Returns:
            :obj:`dict` - stock_information:
                Returns a :obj:`dict` which contains all the values from the introduced Stock in order to create its
                portfolio row.

        """
        stocks = investpy.get_stocks(country=stock.stock_country)

        stock_name = stocks.loc[(stocks['symbol'].str.lower() == stock.stock_symbol.lower()).idxmax(), 'name']
        
        data = investpy.get_stock_historical_data(stock=stock.stock_symbol,
                                                  country=stock.stock_country,
                                                  from_date=stock.purchase_date,
                                                  to_date=date.today().strftime("%d/%m/%Y"))

        currency = data['Currency'][0]

        purchase_cost = self.calculate_purchase_cost(cost_per_share=stock.cost_per_share,
                                                     num_of_shares=stock.num_of_shares)

        current_price = self.get_current_price(data=data)

        gross_current_value = self.calculate_gross_current_value(current_price=current_price,
                                                                 num_of_shares=stock.num_of_shares)

        dividends = investpy.get_stock_dividends(stock=stock.stock_symbol, 
                                                 country=stock.stock_country)

        dividends = dividends.loc[dividends['Payment Date'] < pd.to_datetime(stock.purchase_date, dayfirst=True)].reset_index(drop=True)

        if len(dividends) > 0:
            total_dividends = self.calculate_total_dividends(dividends=dividends,
                                                             num_of_shares=stock.num_of_shares)
        else:
            total_dividends = 0

        net_current_value = self.calculate_net_current_value(gross_current_value=gross_current_value,
                                                             total_dividends=total_dividends)

        total_gain_loss = self.calculate_total_gain_loss(net_current_value=net_current_value,
                                                         purchase_cost=purchase_cost)

        total_gain_loss_percentage = self.calculate_total_gain_loss_percentage(total_gain_loss=total_gain_loss,
                                                                               purchase_cost=purchase_cost)

        info = {
            'stock_symbol': stock.stock_symbol,
            'stock_name': stock_name,
            'stock_country': stock.stock_country,
            'stock_currency': currency,
            'purchase_date': stock.purchase_date,
            'num_of_shares': stock.num_of_shares,
            'cost_per_share': stock.cost_per_share,
            'purchase_cost': purchase_cost,
            'current_price': current_price,
            'gross_current_value': gross_current_value,
            'total_dividends': total_dividends,
            'net_current_value': net_current_value,
            'total_gain_loss': total_gain_loss,
            'total_gain_loss_percentage': total_gain_loss_percentage,
        }

        return info

    def refresh(self):
        """ Method to refresh/reload the StockPortfolio information.
        
        This method is used to refresh or reload the StockPortfolio information since the values may have changed
        since the last time the portfolio was generated. So on, this function will return to get the information
        from every Stock listed in the StockPortfolio.

        """
        if len(self._stock_objs) > 0:
            self._stocks = list()
            for stock_obj in self._stock_objs:
                info = self.__get_stock_info(stock=stock_obj)
                self._stocks.append(info)
            self.data = pd.DataFrame(self._stocks)

    @staticmethod
    def calculate_purchase_cost(cost_per_share, num_of_shares):
        """
        This method calculates the purchase cost, which is the paid value for every stock share hold by the user at
        the time that stock shares were bought.
        """
        return float(cost_per_share * num_of_shares)

    @staticmethod
    def get_current_price(data):
        """
        This method gets the current price value of the introduced stock, which is the last close value indexed in the
        :obj:`pandas.DataFrame`.
        """
        return data.iloc[-1]['Close']

    @staticmethod
    def calculate_gross_current_value(current_price, num_of_shares):
        """
        This method calculates the gross current value which is the total current value of the shares bought,
        which is the result of the multiplication of the current price with the number of bought shares.
        """
        return float(current_price * num_of_shares)
    
    @staticmethod
    def calculate_total_dividends(dividends, num_of_shares):
        """
        This method calculates the total dividend's value which is the sum of all the dividends paid in the time
        range that the user owned/owns the stock shares. Each dividend is paid per the number of stock shares hold
        by the user.
        """
        total_dividends = 0
        for _, dividend in dividends.iterrows():
            total_dividends += (dividend['Dividend'] * num_of_shares)
        return total_dividends

    @staticmethod
    def calculate_net_current_value(gross_current_value, total_dividends):
        """
        This method calculates the net current value which is the total of the gross current value and the value of 
        the dividends, i.e., the total of returns from both capital gains and dividends.
        """
        return gross_current_value + total_dividends

    @staticmethod
    def calculate_total_gain_loss(net_current_value, purchase_cost):
        """
        This method calculates the difference between the current value of the stock shares, including the dividends
        received, with the value that the user paid of them.
        """
        return net_current_value - purchase_cost

    @staticmethod
    def calculate_total_gain_loss_percentage(total_gain_loss, purchase_cost):
        """
        This method compares the stock total gain loss to what the user paid for the stock shares owned and the result
        is a percentage that can show the total stock performance.
        """
        return str(total_gain_loss / purchase_cost) + "%"



# Copyright 2019 Alvaro Bartolome @ alvarobartt in GitHub
# See LICENSE for details.

import pyrtfolio
from pyrtfolio.StockPortfolio import StockPortfolio


def test_package():
    """
    This function tests both the authorship and version of pyrtfolio.
    """

    print(pyrtfolio.__author__)
    print(pyrtfolio.__version__)


def test_stock_portfolio():
    """
    This functions tests the basic creation of a StockPortfolio with some sample Stocks.
    """

    portfolio = StockPortfolio()

    portfolio.add_stock(stock_symbol='BBVA',
                        stock_country='spain',
                        purchase_date='04/01/2018',
                        num_of_shares=2,
                        cost_per_share=7.2)

    portfolio.add_stock(stock_symbol='ELE',
                        stock_country='spain',
                        purchase_date='13/06/2019',
                        num_of_shares=15,
                        cost_per_share=23.8)

    print(portfolio.data.head())
    portfolio.refresh()
    print(portfolio.data.head())


if __name__ == '__main__':
    test_package()
    test_stock_portfolio()



# Copyright 2019 Alvaro Bartolome @ alvarobartt in GitHub
# See LICENSE for details.

__author__ = "Alvaro Bartolome @ alvarobartt in GitHub"
__version__ = "0.4"

from pyrtfolio.StockPortfolio import StockPortfolio



# Copyright 2019 Alvaro Bartolome @ alvarobartt in GitHub
# See LICENSE for details.

import datetime

import investpy


class Stock(object):
    """ Stock is a support class of `pyrtfolio` which validates the Stocks before adding them to StockPortfolio.

    This class is a support class to validate the Stocks before finally adding them to the portfolio created in
    StockPortfolio since the historical data of the introduced Stock needs to be retrieved and, so on, it will error
    if the introduced Stock is not valid.

    Attributes:
        stock_symbol (:obj:`str`): symbol of the Stock that is going to be added to the StockPortfolio.
        stock_country (:obj:`str`): country from where the specified stock_symbol is, so to validate it.
        purchase_date (:obj:`str`):
            date when the shares of the introduced stock were bought, formatted as dd/mm/yyyy.
        num_of_shares (:obj:`int`): amount of shares bought of the specified Stock in the specified date.
        cost_per_share (:obj:`float`): price of every share of the Stock in the specified date.
        valid (:obj:`boolean`): either True or False if the Stock is valid or not, respectively.

    """

    def __init__(self, stock_symbol, stock_country, purchase_date, num_of_shares, cost_per_share):
        """ This is the init method of Stock class which is launched every time the user instances it.

        This method is the init method of this class, Stock, and its main function is to init all the attributes
        contained in it, as previously defined when the function `add_stock()` was called from StockPortfolio. The
        specified values in this class are going to be accessed through the self operator whenever the validation
        function is called, since it is the main function of this class, validating Stocks.

        Args:
            stock_symbol (:obj:`str`): symbol of the Stock that is going to be added to the StockPortfolio.
            stock_country (:obj:`str`): country from where the specified stock_symbol is, so to validate it.
            purchase_date (:obj:`str`):
                date when the shares of the introduced stock were bought, formatted as dd/mm/yyyy.
            num_of_shares (:obj:`int`): amount of shares bought of the specified Stock in the specified date.
            cost_per_share (:obj:`float`): price of every share of the Stock in the specified date.

        """
        self.stock_symbol = stock_symbol
        self.stock_country = stock_country
        self.purchase_date = purchase_date
        self.num_of_shares = num_of_shares
        self.cost_per_share = cost_per_share

        self.valid = False

    def validate(self):
        """ Method used to validate that the introduced Stock is valid before adding it to the StockPortfolio.

        This method is the one in charge of the validation of the introduced Stock via checking the introduced data
        with the one indexed in investpy, so to check if the data match. Also, the introduced parameters are checked in
        order to determine if the type is correct of those values is correct, if not, an exception will be raised. The
        result of this function is just setting either True or False to the self.valid value if the Stock is valid or
        not, respectively.

        """
        if not isinstance(self.stock_symbol, str):
            raise ValueError("ERROR [0005]: The introduced stock_symbol is mandatory and should be a str.")

        if not isinstance(self.stock_country, str):
            raise ValueError("ERROR [0006]: The introduced stock_country is mandatory and should be a str.")

        try:
            datetime.datetime.strptime(self.purchase_date, '%d/%m/%Y')
        except ValueError:
            raise ValueError("ERROR [0007]: The introduced purchase_date is not properly formatted (dd/mm/yyyy).")

        purchase_date_ = datetime.datetime.strptime(self.purchase_date, '%d/%m/%Y')
        if purchase_date_ > datetime.datetime.now():
            raise ValueError("ERROR [0008]: The introduced purchase_date is not valid since it should be earlier than "
                             "the current date.")

        if not isinstance(self.num_of_shares, int):
            raise ValueError("ERROR [0009]: The introduced num_of_shares is mandatory and should be an int higher "
                             "than 0.")
        else:
            if self.num_of_shares <= 0:
                raise ValueError("ERROR [0009]: The introduced num_of_shares is mandatory and should be an int higher "
                                 "than 0.")

        if not isinstance(self.cost_per_share, float):
            raise ValueError("ERROR [0010]: The introduced cost_per_share is mandatory and should be a float higher "
                             "than 0.")
        else:
            if self.cost_per_share <= 0:
                raise ValueError("ERROR [0010]: The introduced Stock is not valid.")

        stock_countries = investpy.get_stock_countries()

        if self.stock_country.lower() in stock_countries:
            stocks = investpy.get_stocks(country=self.stock_country)

            search_results = stocks[stocks['symbol'].str.lower() == self.stock_symbol.lower()]

            if len(search_results) > 0:
                data = investpy.get_stock_historical_data(stock=self.stock_symbol,
                                                          country=self.stock_country,
                                                          from_date=self.purchase_date,
                                                          to_date=datetime.date.today().strftime("%d/%m/%Y"))

                try:
                    purchase_date_ = purchase_date_.strftime("%Y-%m-%d")
                    min_value = data.loc[purchase_date_]['Low']
                    max_value = data.loc[purchase_date_]['High']
                except KeyError:
                    raise KeyError("ERROR [0004]: The introduced purchase_date is not valid since the market was "
                                   "closed.")

                if min_value <= self.cost_per_share <= max_value:
                    self.valid = True
                else:
                    raise ValueError("ERROR [0011]: The introduced value is not possible, because the range stock "
                                     "values of the purchase data were between " + str(min_value) + " and " + str(max_value))
            else:
                raise ValueError("ERROR [0003]: No results were found for the introduced stock_symbol in the specified "
                                 "stock_country.")
        else:
            raise ValueError("ERROR [0002]: The introduced stock_country is not valid or does not have any indexed "
                             "stock.")



# Copyright 2019 Alvaro Bartolome @ alvarobartt in GitHub
# See LICENSE for details.

from setuptools import setup, find_packages
import io


def readme():
    with io.open('README.md', encoding='utf-8') as f:
        return f.read()

def requirements():
    reqs = list()
    with io.open('requirements.txt', encoding='utf-8') as f:
        for line in f.readlines():
            reqs.append(line.strip())
    return reqs


setup(
    name='pyrtfolio',
    version='0.4',
    packages=find_packages(),
    url='https://pyrtfolio.readthedocs.io/',
    download_url='https://github.com/alvarobartt/pyrtfolio/archive/0.4.tar.gz',
    license='GNU General Public License v3 (GPLv3)',
    author='Alvaro Bartolome',
    author_email='alvarobartt@usal.es',
    description='pyrtfolio - Python package to generate stock portfolios',
    long_description=readme(),
    long_description_content_type='text/markdown',
    install_requires=requirements(),
    data_files=[],
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Intended Audience :: Developers",
        "Topic :: Office/Business :: Financial",
        "Topic :: Office/Business :: Financial :: Investment",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    keywords='investing, investing-api, investpy, financial-data, stocks, portfolio',
    python_requires='>=3',
    project_urls={
        'Bug Reports': 'https://github.com/alvarobartt/pyrtfolio/issues',
        'Source': 'https://github.com/alvarobartt/pyrtfolio',
        'Documentation': 'https://pyrtfolio.readthedocs.io/'
    },
)
